// -------------------------------------------------------------
//   Programa que mostra como renderizar um objeto em 3D bem
//   como rotacion-lo em seus eixos
//
//   Polgonos 3D (linhas) - verso 1.0
//                                     
//   Autor   : Stanley A. Sales 
//
// ------------------------------------------------------------- 
//  
//   Nota: Para fins didticos. Portanto, sem otimizao.
//         Assim  possvel ver cada etapa de como so feitos os
//         clculos.
//
// ------------------------------------------------------------- 
//
// Sugestes:
// ==========
//
// Coloque os pontos para seus polgonos em um arquivo e faa a
// leitura do mesmo para que o cdigo no fique extenso.
//
// Otimize o cdigo para desenhar as linhas do polgono usando
// laos de repetio e coordenadas relativas, quando possvel.
//
// Crie outros polgonos
//
// Desenhe mais de um polgono na tela
//
// Movimente horizontal e/ou verticalmente
//
// Pode-se utilizar teclas para mudar o polgino, sua velocidade
// de rotao bem como qual o eixo a ser rotacionado
//
// Mudar a cor de todo o polgono ou somente das arestas
//
// Preencher as faces
//
// Faa polgonos mrficos
//
// -------------------------------------------------------------
// 
Program Poligonos3DLinha;
uses graph;

const
   PI            = 4 * arctan(1);
   FOV           = 180;
   MAX_DISTANCIA = 60;
   MIN_DISTANCIA = 8;

type tipoPontos = record
   x1 : real;              //-
   y1 : real;              //- variveis X, Y e Z
   z1 : real;              //-
   xg : real;              //.
   yg : real;              //. buffers para as equaes
   zg : real;              //.
end;

var
   pontos        : array [0..6, 0..6, 0..6] of tipoPontos;
   sinTable      : array [-360..360] of real;        		// tabela de senos p/ melhorar a performance
   cosTable      : array [-360..360] of real;        		// tabela de cossenos p/ melhorar a performance
   driver        : integer;
   modo          : integer;
   xResolucao    : integer;
   yResolucao    : integer;
   z             : integer;
   xs            : integer;
   ys            : integer;
   zs            : integer;
   t1            : integer;
   t2            : integer;
   t3            : integer;
   dx            : integer;
   dy            : integer;
   dz            : integer;
   j             : integer;
   k             : integer;
   tempo         : integer;
   dist          : real;
   wx1           : real;
   wy1           : real;
   wx2           : real;
   wy2           : real;
   isCubo        : boolean;                              // desenhar cubo?
   isAumentando  : boolean;                              // distncia aumentando?
	    
begin
// Inicializa o modo grfico
   driver := Detect;
   initgraph(driver, modo, '');
  
// Verifica se a inicializao foi feita com sucesso
   if (graphResult <> grOk) then
      begin
         writeln('Erro ao inicializar o modo grfico:', GraphErrorMsg(graphResult));
         exit;
      end;
   
   getAspectRatio(xResolucao, yResolucao);
(*
   Mude esses valores desde que sejam apenas 7 (-3 at 3 so 7 valores)
  
   Se o centro no for zero, o polgono todo vai girar no espao 3D
  
   Esses valores mudam o ponto de rotao do polgono
*)
   for xs := -3 to 3 do                                       // experimente -6 to 0
       begin
          for ys := -3 to 3 do                                // experimente 0 to 6
              begin
                 for zs := -3 to 3 do                         // experimente -1 to 5
                     begin
                        pontos[t1, t2, t3].xg := xs;
                        pontos[t1, t2, t3].yg := ys;
                        pontos[t1, t2, t3].zg := zs;
                        t3                    := t3 + 1;
                     end;
                     
                 t2 := t2 + 1;
                 t3 := 0;
              end;
        
          t1 := t1 + 1;
          t2 := 0;
       end;
(*
   Inicializa as tabelas de seno e cosseno
  
   Calcular seno e cosseno custa caro. Ento  melhor
   calcul-los e guard-los em vetores para futura
   referncia.
  
   Esses vetores so utilizados para rotacionar o objeto
   em um determinado ngulo que vai de -360 a 360 graus.
   Assim, podemos rotacion-lo em qualquer direo.
  
   No podemos usar graus, ento convertemos em radianos
   para que o computador possa trabalhar melhor com eles.
  
   Para converter graus em radianos, basta multiplicar
   graus por PI e dividir por 180.
*)  
	 for z := -360 to 360 do
	     begin                                  
	        sinTable[z] := sin(z * (PI / 180));
	        cosTable[z] := cos(z * (PI / 180));
	     end;
//
// Seed aleatria
//
   randomize;
//
// Inicializa a distncia e cor das linhas
//
   dist := MAX_DISTANCIA;
   SetColor(15);
//
// Loop principal
//
   writeln('Pressione (e mantenha pressionada) uma tecla para finalizar.  :)');
	    
   while not keypressed do
       begin
          dx    := random(13) - 6;          //- velocidade        (-6 a 6)
          dy    := random(13) - 6;          //- de rotao        (-6 a 6)
          dz    := random(13) - 6;          //- de cada eixo      (-6 a 6)
          tempo := random(101) + 50;        // tempo at prximo passo  (50 a 150)
          
					if random(100) mod 2 = 0 then
					   isCubo := not isCubo;          // alterna entre cubo e qualquer outra coisa (no nosso exemplo, pirmide)
					   
          for k := 0 to tempo do
              begin
              
                 if isAumentando then
                    begin
                       dist := dist + 0.5;
                    
                       if dist = MAX_DISTANCIA then
                          isAumentando := false;
                    end
                  else
                    begin
                       dist := dist - 0.5;
                    
                       if dist = MIN_DISTANCIA then
                          isAumentando := true;
                    end;
                
                 for t1 := 0 to 6 do                   //-
                     for t2 := 0 to 6 do               //- rotaciona todos os pontos
                         for t3 := 0 to 6 do           //-
                             begin
                                (*
                                   cada rotao  composta de 2 equaes
                                   cada rotao usa YG, XG ou ZG na equao em vez de Y1, X1 ou Z1
                                   isso se deve ao fato de que, se usssemos Y1, X1 ou Z1, seria usado
                                   novamente na prxima equao e ento teria valor diferente e os
                                   resultados seriam incorretos.
                                   depois das duas equaes, YG, XG ou ZG  atualizada
                                *)
                                
                                // rotaciona o objeto no eixo X. Muda os valores de Y1 e Z1
                                pontos[t1, t2, t3].y1 := pontos[t1, t2, t3].yg * cosTable[dx] - pontos[t1, t2, t3].zg * sinTable[dx];
                                pontos[t1, t2, t3].z1 := pontos[t1, t2, t3].yg * sinTable[dx] + pontos[t1, t2, t3].zg * cosTable[dx];
                                pontos[t1, t2, t3].yg := pontos[t1, t2, t3].y1;
                                pontos[t1, t2, t3].zg := pontos[t1, t2, t3].z1;

								                // rotaciona o objeto no eixo Y. Muda os valores de Z1 e X1
                                pontos[t1, t2, t3].z1 := pontos[t1, t2, t3].zg * cosTable[dy] - pontos[t1, t2, t3].xg * sinTable[dy];
                                pontos[t1, t2, t3].x1 := pontos[t1, t2, t3].zg * sinTable[dy] + pontos[t1, t2, t3].xg * cosTable[dy];
                                pontos[t1, t2, t3].xg := pontos[t1, t2, t3].x1;
    					       	          pontos[t1, t2, t3].zg := pontos[t1, t2, t3].z1;

                                // rotaciona o objeto no eixo Z. Muda os valores de X1 e Y1
                                pontos[t1, t2, t3].x1 := pontos[t1, t2, t3].xg * cosTable[dz] - pontos[t1, t2, t3].yg * sinTable[dz];
                                pontos[t1, t2, t3].y1 := pontos[t1, t2, t3].xg * sinTable[dz] + pontos[t1, t2, t3].yg * cosTable[dz];
                                pontos[t1, t2, t3].xg := pontos[t1, t2, t3].x1;
                                pontos[t1, t2, t3].yg := pontos[t1, t2, t3].y1;
                             end;
                             
                 cleardevice;
(*

 Entender a matriz de pontos 3D  bem simples:
 
 (x, y, z)
  |  |  |
  |  |  +-----> 0 = frente     6 = trs      (Obs.)
  |  +--------> 0 = superior   6 = inferior  (Obs.)
	+-----------> 0 = esquerda   6 = direita
	
	
	Obs.: Dizer se  frente ou no depende do jeito que olhamos.
	      No desenho abaixo, podemos enxergar um cubo com a face frontal
	      em um nvel inferior  face traseira (vendo de cima pra baixo) ou
	      a face frontal em um nvel superior  face traseira (vendo de baixo pra
	      cima). Em uma renderizao mais elaborada, o que manda  o sombreamento
	      (shading). Com o sobreamento, somos forados a enxergar de um s modo.
	      
	      Se enxergarmos, no exemplo abaixo, um cubo visto de cima ento a face
	      frontal estar mais abaixo da traseira e ser a de coordenadas
	      (0, 0, 0), (0, 6, 0), (6, 0, 0) e (6, 6, 0).
	      
	      Se enxergarmos, no exemplo abaixo, um cubo visto de baixo ento a face
	      frontal estar mais acima da traseira e ser a de coordenadas
	      (0, 0, 6), (0, 6, 6), (6, 0, 6) e (6, 6, 6).
	      
	       
                 (0,0,6) --------------------------------------- (6,0,6)
                         |\                                    |\
                         | \                                   | \
                         |  \                                  |  \
                         |   \                                 |   \
                         |    \                                |    \
                         |     \                               |     \
                         |      \                              |      \
                         |       \                             |       \
                         |        \                            |        \
                         | (0,0,0) -------------------------------------- (6,0,0)
                         |         |                           |        |
                         |         |                           |        |
                         |         |                           |        |
                         |         |                           |        |
                         |         |                           |        |
                         |         |                           |        |
                         |         |                           |        |
                         |         |                           |        |
                         |         |                           |        |
                         |         |                           |        |
                 (0,6,6) ----------|---------------------------- (6,6,6)|
                          \        |                           \        |
                           \       |                            \       |
                            \      |                             \      |
                             \     |                              \     |
                              \    |                               \    |
                               \   |                                \   |
                                \  |                                 \  |
                                 \ |                                  \ |
                                  \|                                   \|
                           (0,6,0) -------------------------------------- (6,6,0)

 
 
 Conhecendo isso,  possvel desenhar vrios polgonos, bastando informar os pontos nos quais
 comeam e terminam as linhas!!!
 
 Esses pontos podem estar nos limites externos da matriz ou dentro da mesma.
 
 Nossa matriz tem 7 x 7 x 7 = 7^3 = 343 pontos... d pra desenhar polnogos interessantes.
 
 Por que 7? Ora, de 0 a 6 so 7 valores possveis.
 

Desenhando os pontos na tela
=============================

como desenhar um objeto 3D em uma tela 2D?

temos que converter os pontos X, Y e Z de um espao matemtico
em um plano 2D. Isso  chamado de projeo perspectiva e  bem simples:

X = X * FOV / Z
Y = Y * FOV / Z

onde: FOV = field of view (campo de viso)

adicionamos xResolucao / 2 e yResolucao / 2 a X e Y, respectivamente,
para centralizar na tela

adicionamos dist no Z para aumentar ou diminuir a distncia do observador
(objeto fica menor ou maior na tela)

*)

                 if isCubo then
								    begin                           
								       wx1 := ((pontos[0, 0, 0].x1 * FOV) / (pontos[0, 0, 0].z1 + dist)) + xResolucao / 2;    //
								       wy1 := ((pontos[0, 0, 0].y1 * FOV) / (pontos[0, 0, 0].z1 + dist)) + yResolucao / 2;    //
								       wx2 := ((pontos[6, 0, 0].x1 * FOV) / (pontos[6, 0, 0].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[6, 0, 0].y1 * FOV) / (pontos[6, 0, 0].z1 + dist)) + yResolucao / 2;    //
                                                                                                              //
								       line(int(wx1), int(wy1), int(wx2), int(wy2));                                          //
                                                                                                              //
								       wx2 := ((pontos[6, 6, 0].x1 * FOV) / (pontos[6, 6, 0].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[6, 6, 0].y1 * FOV) / (pontos[6, 6, 0].z1 + dist)) + yResolucao / 2;    //
                                                                                                              //
								       lineTo(int(wx2), int(wy2));                                                            // face frontal (vide obs.)
                                                                                                              //
								       wx2 := ((pontos[0, 6, 0].x1 * FOV) / (pontos[0, 6, 0].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[0, 6, 0].y1 * FOV) / (pontos[0, 6, 0].z1 + dist)) + yResolucao / 2;    //
                                                                                                              //
								       lineTo(int(wx2), int(wy2));                                                            //
                                                                                                              //
								       wx2 := ((pontos[0, 0, 0].x1 * FOV) / (pontos[0, 0, 0].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[0, 0, 0].y1 * FOV) / (pontos[0, 0, 0].z1 + dist)) + yResolucao / 2;    //
                                                                                                              //
								       lineTo(int(wx2), int(wy2));                                                            //

								       wx2 := ((pontos[0, 0, 6].x1 * FOV) / (pontos[0, 0, 6].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[0, 0, 6].y1 * FOV) / (pontos[0, 0, 6].z1 + dist)) + yResolucao / 2;    // aresta superior
																																																				      // esquerda
								       lineTo(int(wx2), int(wy2));                                                            //

								       wx2 := ((pontos[6, 0, 6].x1 * FOV) / (pontos[6, 0, 6].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[6, 0, 6].y1 * FOV) / (pontos[6, 0, 6].z1 + dist)) + yResolucao / 2;    //
                                                                                                              //
								       lineTo(int(wx2), int(wy2));                                                            //
                                                                                                              //
                       wx2 := ((pontos[6, 6, 6].x1 * FOV) / (pontos[6, 6, 6].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[6, 6, 6].y1 * FOV) / (pontos[6, 6, 6].z1 + dist)) + yResolucao / 2;    //
                                                                                                              //
								       lineTo(int(wx2), int(wy2));                                                            // face traseira (vide obs.)
                                                                                                              //
                       wx2 := ((pontos[0, 6, 6].x1 * FOV) / (pontos[0, 6, 6].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[0, 6, 6].y1 * FOV) / (pontos[0, 6, 6].z1 + dist)) + yResolucao / 2;    //
                                                                                                              //
								       lineTo(int(wx2), int(wy2));                                                            //
                                                                                                              //
                       wx2 := ((pontos[0, 0, 6].x1 * FOV) / (pontos[0, 0, 6].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[0, 0, 6].y1 * FOV) / (pontos[0, 0, 6].z1 + dist)) + yResolucao / 2;    //
                                                                                                              //
								       lineTo(int(wx2), int(wy2));                                                            //

								       wx1 := ((pontos[6, 0, 0].x1 * FOV) / (pontos[6, 0, 0].z1 + dist)) + xResolucao / 2;    //
								       wy1 := ((pontos[6, 0, 0].y1 * FOV) / (pontos[6, 0, 0].z1 + dist)) + yResolucao / 2;    //
								       wx2 := ((pontos[6, 0, 6].x1 * FOV) / (pontos[6, 0, 6].z1 + dist)) + xResolucao / 2;    // aresta superior
								       wy2 := ((pontos[6, 0, 6].y1 * FOV) / (pontos[6, 0, 6].z1 + dist)) + yResolucao / 2;    // direita
                                                                                                              //
								       line(int(wx1), int(wy1), int(wx2), int(wy2));                                          //
								 
								       wx1 := ((pontos[6, 6, 0].x1 * FOV) / (pontos[6, 6, 0].z1 + dist)) + xResolucao / 2;    //
								       wy1 := ((pontos[6, 6, 0].y1 * FOV) / (pontos[6, 6, 0].z1 + dist)) + yResolucao / 2;    //
								       wx2 := ((pontos[6, 6, 6].x1 * FOV) / (pontos[6, 6, 6].z1 + dist)) + xResolucao / 2;    // aresta inferior
								       wy2 := ((pontos[6, 6, 6].y1 * FOV) / (pontos[6, 6, 6].z1 + dist)) + yResolucao / 2;    // esquerda
                                                                                                              //
								       line(int(wx1), int(wy1), int(wx2), int(wy2));                                          //
								 
								       wx1 := ((pontos[0, 6, 0].x1 * FOV) / (pontos[0, 6, 0].z1 + dist)) + xResolucao / 2;    //
								       wy1 := ((pontos[0, 6, 0].y1 * FOV) / (pontos[0, 6, 0].z1 + dist)) + yResolucao / 2;    //
								       wx2 := ((pontos[0, 6, 6].x1 * FOV) / (pontos[0, 6, 6].z1 + dist)) + xResolucao / 2;    // aresta inferior
								       wy2 := ((pontos[0, 6, 6].y1 * FOV) / (pontos[0, 6, 6].z1 + dist)) + yResolucao / 2;    // direita
                                                                                                              //
								       line(int(wx1), int(wy1), int(wx2), int(wy2));                                          //
								    end
								 else
								    begin
								       wx1 := ((pontos[0, 6, 0].x1 * FOV) / (pontos[0, 6, 0].z1 + dist)) + xResolucao / 2;    //
								       wy1 := ((pontos[0, 6, 0].y1 * FOV) / (pontos[0, 6, 0].z1 + dist)) + yResolucao / 2;    //
								       wx2 := ((pontos[0, 6, 6].x1 * FOV) / (pontos[0, 6, 6].z1 + dist)) + xResolucao / 2;    // 
								       wy2 := ((pontos[0, 6, 6].y1 * FOV) / (pontos[0, 6, 6].z1 + dist)) + yResolucao / 2;    // 
							      																																									        //
								       line(int(wx1), int(wy1), int(wx2), int(wy2));                                          //
								                                                                                              //
                       wx2 := ((pontos[6, 6, 6].x1 * FOV) / (pontos[6, 6, 6].z1 + dist)) + xResolucao / 2;    //
			      					 wy2 := ((pontos[6, 6, 6].y1 * FOV) / (pontos[6, 6, 6].z1 + dist)) + yResolucao / 2;    //
						      																																													  //
								       lineTo(int(wx2), int(wy2));                                                            // base (face inferior)
                                                                                                              //
                       wx2 := ((pontos[6, 6, 0].x1 * FOV) / (pontos[6, 6, 0].z1 + dist)) + xResolucao / 2;    //
						      		 wy2 := ((pontos[6, 6, 0].y1 * FOV) / (pontos[6, 6, 0].z1 + dist)) + yResolucao / 2;    //
									      																																										  //
      								 lineTo(int(wx2), int(wy2));                                                            //
			      					                                                                                        //
                       wx2 := ((pontos[0, 6, 0].x1 * FOV) / (pontos[0, 6, 0].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[0, 6, 0].y1 * FOV) / (pontos[0, 6, 0].z1 + dist)) + yResolucao / 2;    //
											      																																									//
      								 lineTo(int(wx2), int(wy2));                                                            //
			      					 
                       wx2 := ((pontos[3, 0, 3].x1 * FOV) / (pontos[3, 0, 3].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[3, 0, 3].y1 * FOV) / (pontos[3, 0, 3].z1 + dist)) + yResolucao / 2;    //
      																																																				//
			      					 lineTo(int(wx2), int(wy2));                                                            // um lado (face lateral)
						      																																														//
                       wx2 := ((pontos[6, 6, 0].x1 * FOV) / (pontos[6, 6, 0].z1 + dist)) + xResolucao / 2;    //
			      					 wy2 := ((pontos[6, 6, 0].y1 * FOV) / (pontos[6, 6, 0].z1 + dist)) + yResolucao / 2;    //
						      																																														//
								       lineTo(int(wx2), int(wy2));                                                            //

      								 wx1 := ((pontos[0, 6, 6].x1 * FOV) / (pontos[0, 6, 6].z1 + dist)) + xResolucao / 2;    //
			      					 wy1 := ((pontos[0, 6, 6].y1 * FOV) / (pontos[0, 6, 6].z1 + dist)) + yResolucao / 2;    //
						      		 wx2 := ((pontos[3, 0, 3].x1 * FOV) / (pontos[3, 0, 3].z1 + dist)) + xResolucao / 2;    // outro lado (outra face lateral)
								       wy2 := ((pontos[3, 0, 3].y1 * FOV) / (pontos[3, 0, 3].z1 + dist)) + yResolucao / 2;    // 
											      																																									//
								       line(int(wx1), int(wy1), int(wx2), int(wy2));                                          //
			      
                       wx2 := ((pontos[6, 6, 6].x1 * FOV) / (pontos[6, 6, 6].z1 + dist)) + xResolucao / 2;    //
								       wy2 := ((pontos[6, 6, 6].y1 * FOV) / (pontos[6, 6, 6].z1 + dist)) + yResolucao / 2;    // terceira face lateral
											      																																									//
								       lineTo(int(wx2), int(wy2));                                                            //
								 end;
              end;
       end;

	 closegraph;  
end.